from .movie_review import MovieReview
from .customer_review import CustomerReview
from .imdb import IMDB
from .multi_nli import MultiNLI
from .news_group20 import NewsGroup20
from .reuters_news import ReutersNews
from .ptb import PTB
from .wikitext2 import WikiText2
from .wikitext103 import WikiText103
from .text8 import Text8
from .duc2003 import DUC2003
from .duc2004 import DUC2004
from .gigaword import Gigaword
from .squad import SQuAD



from collections import Counter
import numpy as np
import pandas as pd


class Converter():

    def __init__(self):
        pass

    def flow(self, series):
        return self.to_seq(series)

    def back(self, converted):
        return converted

    def to_seq(self, series):
        seq = series
        if isinstance(seq, pd.Series):
            seq = series.values
        if isinstance(seq, (list, tuple)):
            seq = np.array(seq)
        return seq


class DictionalyConverter(Converter):

    def __init__(self, labels=(), distincts=False):
        super().__init__()
        self.initialize(labels, distincts)

    def initialize(self, labels, distincts):
        self.labels = self.to_seq(labels)
        if not distincts:
            c = Counter(self.labels)
            self.labels = [k_v[0] for k_v in c.most_common()]

    def flow(self, series):
        if len(self.labels) == 0:
            self.initialize(series, False)
        x = [self.labels.index(d) for d in self.to_seq(series)]
        return np.array(x)

    def back(self, converted):
        x = [self.labels[d] for d in converted]
        return np.array(x)


class OneHotConverter(DictionalyConverter):

    def __init__(self, labels, distincts=False, num_class=-1):
        super().__init__(labels, distincts)
        self.num_class = num_class

    def flow(self, series):
        if len(self.labels) == 0:
            self.initialize(series, False)

        x = np.zeros((len(series), max(len(self.labels), self.num_class)))
        for i, d in enumerate(self.to_seq(series)):
            _id = self.labels.index(d)
            x[i][_id] = 1
        return x

    def back(self, converted):
        ids = np.argmax(converted, axis=1)
        x = [self.labels[d] for d in ids]
        return np.array(x)


class VocabularyConverter(Converter):

    def __init__(self, vocabulary, fixed_len=-1):
        self._vocab = vocabulary
        self.fixed_len = fixed_len if fixed_len > 0 else vocabulary.max_len

    def flow(self, series):
        x = [self._vocab.str_to_matrix(d, fixed_len=self.fixed_len)
             for d in self.to_seq(series)]
        return np.array(x)

    def back(self, converted):
        x = [self._vocab.matrix_to_words(d) for d in converted]
        return np.array(x)



import os
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class IMDB(Dataset):

    def __init__(self):
        super().__init__(
            name="Large Movie Review Dataset(IMDB)",
            site_url="http://ai.stanford.edu/~amaas/data/sentiment/",
            download_url="http://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz",  # noqa
            description="Movie review data is constructed by 25,000 reviews " \
                        "that have positive/negative annotation"
            )

    def download(self,
                 directory="", shuffle=True, test_size=0, sample_count=0,
                 force=False):
        if test_size != 0:
            raise Exception("The dataset is already splitted to train & test.")

        return super().download(directory, shuffle, 0, sample_count, force)

    def prepare(self, dataset_root, extracted_path):
        extracted_dir = os.path.join(extracted_path, "aclImdb")
        data_dirs = ["train", "test"]
        pathes = []
        for d in data_dirs:
            target_dir = os.path.join(extracted_dir, d)
            file_path = os.path.join(dataset_root, "imdb_" + d + ".txt")
            self.label_by_dir(
                file_path, target_dir, {"pos": 1, "neg": 0}, task_size=1000)

            pathes.append(file_path)

            if d == "train":
                unlabeled = os.path.join(dataset_root, "imdb_unlabeled.txt")
                self.label_by_dir(
                    unlabeled, target_dir, {"unsup": None}, task_size=1000)
                pathes.append(unlabeled)

        return pathes[0]

    def make_resource(self, data_root):
        return IMDBResource(data_root)

    @classmethod
    def _parallel_parser(cls, label, path):
        features = cls._file_to_features(path)
        if label is not None:
            line = "\t".join([str(label)] + features) + "\n"
        else:
            line = "\t".join(features) + "\n"  # unlabeled
        return line

    @classmethod
    def _file_to_features(cls, path):
        # override this method if you want implements custome process
        file_name = os.path.basename(path)
        f, ext = os.path.splitext(file_name)
        els = f.split("_")
        rating = 0
        if len(els) == 2:
            rating = els[-1]

        review = ""
        with open(path, encoding="utf-8") as f:
            lines = f.readlines()
            lines = [ln.replace("\t", " ").strip() for ln in lines]
            review = " ".join(lines)

        if rating != "0":
            return [rating, review]
        else:
            return [review]


class IMDBResource(Resource):

    def __init__(self,
                 root,
                 columns=None, target="",
                 separator="\t", pattern=()):

        super().__init__(
            root,
            ["polarity", "rating", "review"],
            "polarity",
            separator,
            {
                "train": "_train",
                "test": "_test",
                "valid": "_valid",
                "unlabeled": "_unlabeled",
                "sample": "_samples"
            })

    @property
    def unlabeled_file_path(self):
        return self._get_prop("unlabeled")

    def unlabeled_data(self, split_target=False):
        return self._get_data("unlabeled", split_target)



import os
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class Text8(Dataset):

    def __init__(self, kind="en"):
        super().__init__(
            name="Text8",
            site_url="http://mattmahoney.net/dc/textdata",
            download_url="http://mattmahoney.net/dc/text8.zip",
            description="This dataset offers cleaned English Wikipedia text"
            )
        self.kind = kind
        if self.kind == "ja":
            self.site_url = "https://github.com/Hironsan/ja.text8"
            self.download_url = "https://s3-ap-northeast-1.amazonaws.com/dev.tech-sketch.jp/chakki/public/ja.text8.zip"  # noqa
            self.description = self.description.replace("English", "Japanese")

    @classmethod
    def en(cls):
        return Text8("en")

    @classmethod
    def ja(cls):
        return Text8("ja")

    def download(self,
                 directory="", shuffle=True, test_size=10, sample_count=0,
                 force=False):
        if test_size < 1:
            raise Exception("This dataset is splitted by byte. test_size "
                            "means how many mega bytes are used for test.")

        self._test_size = test_size
        return super().download(directory, False, 0, sample_count, force)

    @property
    def extract_targets(self):
        inner_file = "text8" if self.kind == "en" else "ja.text8"
        return [inner_file]

    def prepare(self, dataset_root, extracted_path):
        inner_file = self.extract_targets[0]
        target = "text8" if self.kind == "en" else "text8ja"
        train_file_path = os.path.join(dataset_root, target + ".train.txt")
        if self._test_size == 0:
            os.rename(os.path.join(dataset_root, inner_file), train_file_path)
        else:
            test_file_path = os.path.join(dataset_root, target + ".test.txt")

            line = ""
            test_byte = self._test_size * 1000000
            with open(os.path.join(extracted_path, inner_file),
                      encoding="utf-8") as f:
                line = f.readline().strip()
            with open(train_file_path, mode="w", encoding="utf-8") as train:
                train.write(line[:-test_byte])
            with open(test_file_path, mode="w", encoding="utf-8") as test:
                test.write(line[-test_byte:])

        return train_file_path

    def make_resource(self, data_root):
        return Resource(
            data_root,
            ["sentence"],
            pattern={
                "train": ".train",
                "test": ".test",
                "valid": ".valid",
                "samples": "_samples"
            }
        )



import os
import sys
import re
import mmap
import random
import math
import zipfile
import tarfile
import gzip
import shutil
from urllib.parse import urlparse
import requests
from chazutsu.datasets.framework.xtqdm import xtqdm
from joblib import Parallel, delayed
from chazutsu.datasets.framework.resource import Resource


class Dataset():
    """ Dataset Class Framework """

    def __init__(self,
                 name, site_url, download_url, description,
                 log_level=None, test_download_url=""):
        """
        Args:
            name        : dataset name
            site_url    : url to access the dataset homepage
            download_url    : dataset download url
            description     : description of dataset
        """

        self.name = name
        self.site_url = site_url
        self.download_url = download_url
        self.description = description
        # When test_mode = True, use the test_url
        self.test_mode = False
        self.test_download_url = test_download_url
        if not self.test_download_url:
            self.test_download_url = download_url
        self.__trush = []

        # create logger
        from logging import getLogger, StreamHandler, DEBUG
        self.logger = getLogger(self.__class__.__name__.lower())
        if not self.logger.hasHandlers():
            # logger is global object!
            _level = DEBUG if log_level is None else log_level
            handler = StreamHandler(sys.stdout)
            handler.setLevel(_level)
            self.logger.setLevel(_level)
            self.logger.addHandler(handler)

    @property
    def root_name(self):
        return self.name.lower().replace(" ", "_")
    
    @property
    def extract_targets(self):
        return ()

    def on_test(self):
        self.test_mode = True
        return self

    def get_dataset_root(self, directory):
        return os.path.join(directory, self.root_name)

    def get_extracted_path(self, compressed_file):
        dataset_root, file_name = os.path.split(compressed_file)
        return os.path.join(dataset_root, "_extracted")

    def save_and_extract(self, directory="", force=False):
        _dir = self.check_directory(directory)
        dataset_root = self.get_dataset_root(_dir)
        if not os.path.isdir(dataset_root):
            os.mkdir(dataset_root)

        # download and save file
        save_file_path = self.save_dataset(dataset_root)

        # extract dataset file from saved file
        extracted_file_path = self.extract(save_file_path)

        self.trush(save_file_path)
        self.trush(extracted_file_path)
        return dataset_root, extracted_file_path

    def download(self,
                 directory="", shuffle=True, test_size=0.3, sample_count=0,
                 force=False):
        _dir = self.check_directory(directory)
        dataset_root = self.get_dataset_root(_dir)
        save_file_path = os.path.join(dataset_root, self._get_file_name(None))
        if os.path.isdir(dataset_root) and not os.path.exists(save_file_path) \
           and (not force and not self.test_mode):
            r = self.make_resource(dataset_root)
            if len(r._resource) > 0:
                # data_root already exists and have contents
                self.logger.info(
                    "Read resource from the existed resource"
                    "(if you want to retry, set force=True).")
                return self.make_resource(dataset_root)

        dataset_root, extracted_path = self.save_and_extract(
            directory, force)

        prepared_file_path = self.prepare(dataset_root, extracted_path)

        if not self.test_mode:
            self.clear_trush()

        if shuffle:
            self.logger.info("Shuffle the extracted dataset.")
            lines = []
            with open(prepared_file_path, encoding="utf-8") as f:
                lines = f.readlines()
            random.shuffle(lines)
            with open(prepared_file_path, "w", encoding="utf-8") as f:
                f.writelines(lines)

        # make sample file
        self.make_samples(prepared_file_path, sample_count)

        # split to train & test
        self.train_test_split(prepared_file_path, test_size)

        if not self.test_mode:
            self.clear_trush()

        self.logger.info(
            "Done all process! Make below files at {}".format(dataset_root))
        for f in os.listdir(dataset_root):
            if f.startswith("."):
                continue
            self.logger.info(" " + f)

        r = self.make_resource(dataset_root)

        return r

    def load(self, directory=""):
        dir = self.check_directory(directory)
        dataset_root = os.path.join(dir, self._get_root_name())
        if os.path.isdir(dataset_root):
            return self.make_resource(dataset_root)
        else:
            return None

    def check_directory(self, directory):
        if os.path.isdir(directory):
            return directory
        elif len(directory) > 0:
            os.mkdir(directory)
            return directory
        else:
            current = os.getcwd()
            data_dir = os.path.join(current, "data")
            if not os.path.isdir(data_dir):
                self.logger.info(
                    "Make directory for download at {}.".format(current))
                os.mkdir(data_dir)
            return data_dir

    def save_dataset(self, dataset_root):
        save_file_path = os.path.join(dataset_root, self._get_file_name(None))
        if os.path.exists(save_file_path):
            self.logger.info("The dataset file already exists.")
            return save_file_path

        url = self.test_download_url if self.test_mode else self.download_url
        # download and save it as raw file
        self.logger.info(
            "Begin downloading the {} dataset from {}.".format(self.name, url))
        resp = requests.get(self.download_url, stream=True)
        if not resp.ok:
            raise Exception("Can not get dataset from {}.".format(url))

        # save content in response to file
        total_size = int(resp.headers.get("content-length", 0))
        file_name = self._get_file_name(resp)
        _, ext = os.path.splitext(file_name)
        save_file_path = os.path.abspath(os.path.join(dataset_root, file_name))
        self.logger.info("The dataset is saved to {}".format(save_file_path))
        with open(save_file_path, "wb") as f:
            chunk_size = 1024
            limit = total_size / chunk_size
            for data in xtqdm(resp.iter_content(chunk_size=chunk_size),
                              total=limit, unit="B", unit_scale=True):
                f.write(data)

        return save_file_path

    def extract(self, compressed_file):
        target = self.get_extracted_path(compressed_file)
        if len(self.extract_targets) == 0:
            self.extractall(compressed_file)
        else:
            self.extract_file(compressed_file, self.extract_targets)

        return target

    def extractall(self, compressed_file):
        target = self.get_extracted_path(compressed_file)
        if os.path.exists(target):
            self.logger.info("The file already expanded.")
            return target
        else:
            os.mkdir(target)

        if zipfile.is_zipfile(compressed_file):
            with zipfile.ZipFile(compressed_file) as z:
                z.extractall(path=target)
        elif tarfile.is_tarfile(compressed_file):
            with tarfile.open(compressed_file) as t:
                t.extractall(path=target)
        elif compressed_file.endswith(".gz"):
            os.mkdir(target)
            with gzip.open(compressed_file, "rb") as g:
                file_name = os.path.basename(compressed_file)
                file_base, _ = os.path.splitext(file_name)
                p = os.path.join(target, file_base)
                with open(p, "wb") as f:
                    for ln in g:
                        f.write(ln)

        self.trush(compressed_file)

        return target

    def extract_file(self, compressed_file, relative_pathes):
        # unpack the archive file and extract directed path file
        base = self.get_extracted_path(compressed_file)
        if os.path.exists(base):
            self.logger.info("The file already expanded.")
            return base
        else:
            os.mkdir(base)
        is_path_lists = isinstance(relative_pathes, (tuple, list))
        target = relative_pathes if is_path_lists else [relative_pathes]

        extracteds = []
        if zipfile.is_zipfile(compressed_file):
            with zipfile.ZipFile(compressed_file) as z:
                for n in filter(lambda n: n in target, z.namelist()):
                    file_name = os.path.basename(n)
                    p = os.path.join(base, file_name)
                    with open(p, "wb") as f:
                        f.write(z.read(n))
                    extracteds.append(p)

        elif tarfile.is_tarfile(compressed_file):
            with tarfile.open(compressed_file) as t:
                for m in filter(lambda m: m in target, t.getnames()):
                    file_name = os.path.basename(m)
                    p = os.path.join(base, file_name)
                    with open(p, "wb") as f:
                        with t.extractfile(m) as tf:
                            for c in tf:
                                f.write(c)
                    extracteds.append(p)

        elif compressed_file.endswith(".gz"):
            with gzip.open(compressed_file, "rb") as g:
                p = os.path.join(base, relative_pathes[0])
                extracteds.append(p)
                with open(p, "wb") as f:
                    for ln in g:
                        f.write(ln)

        # remove downloaded raw file (zip/tar.gz etc)
        self.trush(compressed_file)

        return extracteds
    
    def prepare(self, dataset_root, extracted_path):
        return extracted_path

    def move_extracteds(self, dataset_root, extracted_path, filter=""):
        moveds = []
        for e in self.extract_targets:
            if filter and filter not in e:
                continue
            file_name = os.path.basename(e)
            s = os.path.join(extracted_path, file_name)
            t = os.path.join(dataset_root, file_name)
            shutil.move(s, t)
            moveds.append(t)
        return moveds

    def label_by_dir(self, file_path, target_dir, dir_and_label, task_size=10):
        label_dirs = dir_and_label.keys()
        dirs = [d for d in os.listdir(target_dir)
                if os.path.isdir(os.path.join(target_dir, d))
                and d in label_dirs]

        write_flg = True
        for d in dirs:
            self.logger.info(
                "Extracting {} (labeled by {}).".format(d, dir_and_label[d]))
            label = dir_and_label[d]
            dir_path = os.path.join(target_dir, d)
            pathes = [os.path.join(dir_path, f) for f in os.listdir(dir_path)]
            pathes = [p for p in pathes if os.path.isfile(p)]
            task_length = int(math.ceil(len(pathes) / task_size))
            for i in xtqdm(range(task_length)):
                index = i * task_size
                tasks = pathes[index:(index + task_size)]
                lines = Parallel(n_jobs=-1)(
                        delayed(self._make_pair)(label, t) for t in tasks)
                mode = "w" if write_flg else "a"
                with open(file_path, mode=mode, encoding="utf-8") as f:
                    for ln in lines:
                        f.write(ln)
                write_flg = False
    
    @classmethod
    def _make_pair(cls, label, path):
        features = cls._file_to_features(path)
        line = "\t".join([str(label)] + features) + "\n"
        return line

    @classmethod
    def _file_to_features(cls, path):
        # override this method if you want implements custome process
        fs = []
        with open(path, encoding="utf-8") as f:
            lines = f.readlines()
            lines = [ln.replace("\t", " ").strip() for ln in lines]
            fs = [" ".join(lines)]
        return fs

    def train_test_split(self, original_file_path, test_size):
        if test_size < 0 or test_size > 1:
            self.logger.error(
                "test_size have to be between 0 ~ 1." \
                "if you don't want to split, please set 0.")
            return []
        elif test_size == 0 or test_size == 1:
            return []

        self.logger.info("Split to train & test file.")

        total_count = self.get_line_count(original_file_path)
        test_count = int(round(total_count * test_size))
        test_targets = random.sample(range(total_count), test_count)
        
        base, ext = os.path.splitext(original_file_path)
        train_test_path = [base + x + ext for x in ["_train", "_test"]]
        train_file = open(train_test_path[0], "wb")
        test_file = open(train_test_path[1], "wb")

        with open(original_file_path, "rb") as f:
            i = 0
            for line in xtqdm(f, total=total_count):
                target = test_file if i in test_targets else train_file
                target.write(line)
                i += 1

        train_file.close()
        test_file.close()

        self.logger.info(
            "Train & Test file is {}({}rows) & {}({}rows, {:.2f}%).".format(
             os.path.basename(train_test_path[0]), total_count - test_count,
             os.path.basename(train_test_path[1]), test_count, 
             test_count / total_count * 100)
            )

        self.trush(original_file_path)

        return train_test_path

    def make_samples(self, original_file_path, sample_count):
        if sample_count == 0:
            return ""
        
        base, ext = os.path.splitext(original_file_path)
        sample_path = base + "_samples" + ext
        samples_file = open(sample_path, "wb")

        total_count = self.get_line_count(original_file_path)
        # for reproducibility of sampling, use fixed interval
        sample_target = range(0, total_count, total_count // sample_count)
        with open(original_file_path, "rb") as f:
            count = 0
            i = 0
            for line in f:
                if i in sample_target:
                    samples_file.write(line)
                    count += 1
                i += 1
                if count == sample_count:
                    break

        samples_file.close()
        self.logger.info("Sample file is {} (picking {} samples).".format(
            os.path.basename(sample_path), sample_count)
        )

        return sample_path

    def make_resource(self, data_root):
        return Resource(data_root)

    def _get_file_name(self, resp):
        file_name = ""
        if resp and "content-disposition" in resp.headers:
            cd = resp.headers["content-disposition"]
            file_matches = re.search("filename=(.+)", cd)
            if file_matches:
                file_name = file_matches.group(0)
                file_name = file_name.split("=")[1]
        else:
            parsed = urlparse(self.download_url)
            file_name = os.path.basename(parsed.path)

        if self.test_mode:
            file_name = "test_" + file_name

        return file_name

    def get_line_count(self, file_path):
        count = 0
        with open(file_path, "r+") as fp:
            buf = mmap.mmap(fp.fileno(), 0)
            while buf.readline():
                count += 1
        return count

    def trush(self, path):
        self.__trush.append(path)

    def clear_trush(self):
        for t in self.__trush:
            if not os.path.exists(t):
                continue
            else:
                if os.path.isfile(t):
                    os.remove(t)
                if os.path.isdir(t):
                    shutil.rmtree(t)

    def show(self):
        print("About {}".format(self.name))
        print(self.description)
        print("see also: {}".format(self.site_url))



import os
import shutil
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class PTB(Dataset):

    def __init__(self):
        super().__init__(
            name="Penn Tree Bank",
            site_url="https://github.com/tomsercu/lstm",
            download_url="https://s3-ap-northeast-1.amazonaws.com/dev.tech-sketch.jp/chakki/chazutsu/ptb.zip",  # noqa
            description="basic language modeling dataset that omits linguistic structure annotations."  # noqa
            )

    @property
    def extract_targets(self):
        return ["ptb/ptb.train.txt", "ptb/ptb.valid.txt", "ptb/ptb.test.txt"]

    def download(self,
                 directory="", shuffle=True, test_size=0, sample_count=0,
                 force=False):
        if test_size != 0:
            raise Exception("The dataset is already splitted to train & test.")
        # in language modeling, shuffle is not needed
        return super().download(directory, False, 0, sample_count, force)

    def prepare(self, dataset_root, extracted_path):
        self.move_extracteds(dataset_root, extracted_path)
        train_file_path = os.path.join(dataset_root, "ptb.train.txt")
        return train_file_path

    def make_resource(self, data_root):
        return Resource(
            data_root,
            ["sentence"],
            pattern={
                "train": ".train",
                "test": ".test",
                "valid": ".valid",
                "samples": "_samples"
            }
        )



import os
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class WikiText2(Dataset):

    def __init__(self):
        super().__init__(
            name="WikiText-2",
            site_url="https://metamind.io/research/the-wikitext-long-term-dependency-language-modeling-dataset/",  # noqa
            download_url="https://s3.amazonaws.com/research.metamind.io/wikitext/wikitext-2-v1.zip",  # noqa
            description="The dataset for language modeling that is larger than PTB (over 2 times)."  # noqa
            )

    def download(self,
                 directory="", shuffle=True, test_size=0, sample_count=0,
                 force=False):
        if test_size != 0:
            raise Exception("The dataset is already splitted to train & test.")

        # in language modeling, shuffle is not needed
        return super().download(directory, False, 0, sample_count, force)

    @property
    def extract_targets(self):
        return [
                "wikitext-2/wiki.train.tokens",
                "wikitext-2/wiki.test.tokens",
                "wikitext-2/wiki.valid.tokens"
                ]

    def prepare(self, dataset_root, extracted_path):
        self.move_extracteds(dataset_root, extracted_path)
        train_file_path = os.path.join(dataset_root, "wiki.train.tokens")
        return train_file_path

    def make_resource(self, data_root):
        return Resource(
            data_root,
            ["sentence"],
            pattern={
                "train": ".train",
                "test": ".test",
                "valid": ".valid",
                "samples": "_samples"
            }
        )



import os
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class Gigaword(Dataset):

    def __init__(self):

        super().__init__(
            name="Gigaword",
            site_url="https://catalog.ldc.upenn.edu/ldc2003t05",
            download_url="https://s3-ap-northeast-1.amazonaws.com/dev.tech-sketch.jp/chakki/chazutsu/Gigaword.zip",  # noqa
            description="Gigaword is news text corpora con. " \
                        "This dataset is preprocessed version that is used at " \
                        "'A Neural Attention Model for Sentence Summarization'."
            )

    @property
    def extract_targets(self):
        return ["Gigaword/input.txt",
                "Gigaword/task1_ref0.txt"]

    def prepare(self, dataset_root, extracted_path):
        news_path = self.__get_extracted_path(extracted_path, 0)
        summary_path = self.__get_extracted_path(extracted_path, 1)
        write_file_name = "gigaword.txt"
        write_path = os.path.join(dataset_root, write_file_name)

        write_file = open(write_path, "w", encoding="utf-8")
        with open(news_path, encoding="utf-8") as n:
            with open(summary_path, encoding="utf-8") as s:
                for nline, sline in zip(n, s):
                    line = "\t".join([
                        nline.strip(), sline.strip()
                        ]) + "\n"
                    write_file.write(line)
        write_file.close()
        return write_path

    def __get_extracted_path(self, extracted_path, index):
        return os.path.join(
                extracted_path, 
                os.path.basename(self.extract_targets[index]))

    def make_resource(self, data_root):
        return Resource(
                data_root,
                columns=["news", "summary"],
                target="summary")



import os
import csv
import json
from collections import Counter

from chazutsu.datasets.framework.xtqdm import xtqdm
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class SQuAD(Dataset):

    def __init__(self, kind="train", version="v2.0"):

        super().__init__(
            name="SQuAD",
            site_url="https://rajpurkar.github.io/SQuAD-explorer/",
            download_url="",
            description="Stanford Question Answering Dataset (SQuAD) is a new reading comprehension dataset,"
            "consisting of questions posed by crowdworkers on a set of Wikipedia articles, "
            "where the answer to every question is a segment of text, or span, from the corresponding reading passage."
            "With 100,000+ question-answer pairs on 500+ articles,"
            "SQuAD is significantly larger than previous reading comprehension datasets."
        )

        versions = ("v1.1", "v2.0")

        if version not in ("v1.1", "v2.0"):
            raise Exception(
                "You have to choose version from {}".format(",".join(versions)))

        endpoint = "https://rajpurkar.github.io/SQuAD-explorer/dataset/"
        urls = {
            "train": endpoint + "train-{}.json".format(version),
            "dev": endpoint + "dev-{}.json".format(version)
        }

        if kind not in urls:
            keys = ",".join(urls.keys())
            raise Exception("You have to choose kind from {}".format(keys))

        self.kind = kind
        self.version = version
        self.download_url = urls[kind]
        self.original_file = os.path.basename(self.download_url)
        self.columns = ["context", "question", "answer", "start", "end"]

    @classmethod
    def train(cls):
        return cls("train")

    @classmethod
    def dev(cls):
        return cls("dev")

    @property
    def root_name(self):
        return self.name.lower().replace(" ", "_") + "_{}_{}".format(self.kind, self.version)

    def download(self,
                 directory="", shuffle=True, test_size=0, sample_count=0,
                 force=False):
        if test_size:
            raise Exception("The dataset is already splitted to train & dev.")

        return super().download(directory, shuffle, 0, sample_count, force)

    def prepare(self, dataset_root, _):
        original_file_path = os.path.join(dataset_root, self.original_file)
        write_file_path = os.path.splitext(original_file_path)[0] + ".txt"
        write_file = open(write_file_path, mode="w", encoding="utf-8", newline="")
        writer = csv.writer(write_file, delimiter="\t")

        self.logger.info("Preprocessing {}".format(original_file_path))
        with open(original_file_path, encoding="utf-8") as rf:
            data = json.load(rf)["data"]

        make_row = getattr(
            self, "make_row_{}".format(self.version.replace(".", "_")))

        for article in xtqdm(data):
            for paragraph in article["paragraphs"]:
                context = paragraph["context"].replace("\n", " ")
                for qa in paragraph["qas"]:
                    question = qa["question"].strip().replace("\n", "")
                    row = make_row(context, question, qa)
                    writer.writerow(row)

        self.trush(original_file_path)
        write_file.close()

        return write_file_path

    @staticmethod
    def make_row_v1_1(context, question, qa):
        answers, starts = zip(
            *((a["text"], a["answer_start"]) for a in qa["answers"]))
        spans = [(start, start + len(text))
                 for text, start in zip(answers, starts)]
        most_frequent_span = Counter(spans).most_common(1)[0][0]
        target_index = max(range(len(spans)),
                           key=lambda x: spans[x] == most_frequent_span)
        # pick the most frequent answer
        answer = answers[target_index]
        start, end = most_frequent_span
        return (context, question, start, end, answer)

    @classmethod
    def make_row_v2_0(cls, context, question, qa):
        if not qa["is_impossible"]:
            # You can answer the question
            return cls.make_row_v1_1(context, question, qa)
        else:
            # You can't answer the question
            return (context, question, -1, -1, "")

    def make_resource(self, data_root):
        if self.kind == "train":
            return Resource(data_root, columns=self.columns)
        elif self.kind == "dev":
            return Resource(data_root, columns=self.columns)



import os
import re
import json
from chazutsu.datasets.framework.xtqdm import xtqdm
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class MultiNLI(Dataset):
    LABEL_MAP = {
        "entailment": 0,
        "neutral": 1,
        "contradiction": 2,
        "hidden": 0
    }

    def __init__(self, matched=True, full=False):
        """
        matched: matched dataset for mismatched dataset
        full: use full columns in the dataset. If full=False then extract "
              only "gold_label", "genre", "sentence1", "sentence2"
        """
        super().__init__(
            name="MultiNLI",
            site_url="http://www.nyu.edu/projects/bowman/multinli/",
            download_url="https://s3-ap-northeast-1.amazonaws.com/dev.tech-sketch.jp/chakki/chazutsu/multinli_0.9_jsonl.zip",  # noqa
            description="multi genre's text entailment dataset. " \
                        "There are in-domain(matched) and " \
                        "cross-domain(mismatched) data."
            )
        self.matched = matched
        self.is_full = full
        self.columns = []
        if self.is_full:
            self.columns = [
                "label", "annotator_labels", "genre", "pairID",
                "sentence1", "sentence2", "sentence1_parse", "sentence2_parse"]
        else:
            self.columns = [
                "label", "genre", "pairID", "sentence1", "sentence2"]
        self._tokenize_pattern = re.compile("\(|\)")

    def download(self,
                 directory="", shuffle=True, test_size=0, sample_count=0,
                 force=False):
        if test_size != 0:
            raise Exception("The dataset is already splitted to train & test.")
        # in language modeling, shuffle is not needed
        return super().download(directory, shuffle, 0, sample_count, force)

    @classmethod
    def matched(cls, full=False):
        return MultiNLI(True, full)

    @classmethod
    def mismatched(cls, full=False):
        return MultiNLI(False, full)

    @property
    def kind(self):
        return "matched" if self.matched else "mismatched"

    @property
    def root_name(self):
        return self.name.lower().replace(" ", "_") + "_" + self.kind

    @property
    def extract_targets(self):
        return [
                "multinli_0.9_jsonl/multinli_0.9_train.jsonl",
                "multinli_0.9_jsonl/multinli_0.9_{}_dev.jsonl".format(self.kind),  # noqa
                "multinli_0.9_jsonl/multinli_0.9_{}_unlabeled_test.jsonl".format(self.kind)  # noqa
            ]

    def prepare(self, dataset_root, extracted_path):
        train_file = ""
        for e in self.extract_targets:
            p = os.path.join(extracted_path, os.path.basename(e))
            preprocessed = self.preprocess_file(dataset_root, p)
            if "_train" in preprocessed:
                train_file = preprocessed
            self.trush(p)

        return train_file

    def preprocess_file(self, dataset_root, path):
        write_file_name = os.path.basename(path).replace(".jsonl", ".txt")
        write_file_path = os.path.join(dataset_root, write_file_name)
        write_file = open(write_file_path, mode="w", encoding="utf-8")
        file_kind = path.split("_")[-1]

        self.logger.info("Preprocessing {} file".format(file_kind))
        total_count = self.get_line_count(path)
        with open(path, encoding="utf-8") as rf:
            for line in xtqdm(rf, total=total_count):
                preprocessed = self.preprocess_jsonl(line)
                if preprocessed:
                    w_line = "\t".join(preprocessed) + "\n"
                    write_file.write(w_line)

        write_file.close()
        return write_file_path

    def preprocess_jsonl(self, jsonl):
        loaded = json.loads(jsonl)
        if loaded["gold_label"] not in self.LABEL_MAP:
            return None

        loaded["label"] = str(self.LABEL_MAP[loaded["gold_label"]])
        if "genre" not in loaded or not loaded["genre"]:
            loaded["genre"] = "-"
        for sentence in ["sentence1", "sentence2"]:
            binary = sentence + "_binary_parse"
            loaded[sentence] = self._tokenized_str(loaded[binary])

        values = [loaded[c] for c in self.columns]
        return values

    def _tokenized_str(self, sentence):
        _s = self._tokenize_pattern.sub("", sentence)
        tokens = _s.split()
        tokens = [t.strip() for t in tokens if t.strip()]
        return " ".join(tokens)

    def make_resource(self, data_root):
        return NLIResource(data_root, columns=self.columns, target="label")


class NLIResource(Resource):

    def __init__(self, root,
                 columns=None, target="",
                 separator="\t", pattern=()):

        super().__init__(
            root,
            columns,
            target,
            separator,
            {
                "train": "_train",
                "test": "_test",
                "dev": "_dev",
                "sample": "_samples"
            })

    @property
    def dev_file_path(self):
        return self._get_prop("dev")

    def dev_data(self, split_target=False):
        return self._get_data("dev", split_target)



import os
from chazutsu.datasets.framework.xtqdm import xtqdm
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class MovieReview(Dataset):

    def __init__(self, kind="polarity"):
        super().__init__(
            name="Moview Review Data",
            site_url="http://www.cs.cornell.edu/people/pabo/movie-review-data/",  # noqa
            download_url="",
            description="movie review data is annotated by 3 kinds of label" \
            " (polarity, subjective rating, subjectivity)."
            )

        urls = {
            "polarity": "http://www.cs.cornell.edu/people/pabo/movie-review-data/review_polarity.tar.gz",  # noqa
            "polarity_v1": "http://www.cs.cornell.edu/people/pabo/movie-review-data/rt-polaritydata.tar.gz",  # noqa
            "rating": "http://www.cs.cornell.edu/people/pabo/movie-review-data/scale_data.tar.gz",  # noqa
            "subjectivity": "http://www.cs.cornell.edu/people/pabo/movie-review-data/rotten_imdb.tar.gz"  # noqa
        }

        if kind not in urls:
            keys = ",".join(urls.keys())
            raise Exception("You have to choose kind from {}".format(keys))

        self.kind = kind
        self.download_url = urls[self.kind]

    @classmethod
    def polarity(cls):
        return MovieReview("polarity")

    @classmethod
    def polarity_v1(cls):
        return MovieReview("polarity_v1")

    @classmethod
    def rating(cls):
        return MovieReview("rating")

    @classmethod
    def subjectivity(cls):
        return MovieReview("subjectivity")

    @property
    def root_name(self):
        return self.name.lower().replace(" ", "_") + "_" + self.kind

    @property
    def extract_targets(self):
        if self.kind == "polarity_v1":
            return ["rt-polaritydata/rt-polarity.neg",
                    "rt-polaritydata/rt-polarity.pos"]
        elif self.kind == "subjectivity":
            return ["plot.tok.gt9.5000",
                    "quote.tok.gt9.5000"]
        return ()

    def prepare(self, dataset_root, extracted_path):
        if self.kind == "polarity":
            return self._prepare_polarity(dataset_root, extracted_path)
        elif self.kind == "polarity_v1":
            return self._prepare_polarity_v1(dataset_root, extracted_path)
        elif self.kind == "rating":
            return self._prepare_rating(dataset_root, extracted_path)
        elif self.kind == "subjectivity":
            return self._prepare_subjectivity(dataset_root, extracted_path)
        else:
            raise Exception(
                "{} is not supported in extraction process.".format(self.kind))

    def make_resource(self, data_root):
        if self.kind in ["polarity", "polarity_v1"]:
            return Resource(data_root,
                            columns=["polarity", "review"], target="polarity")
        elif self.kind == "rating":
            return Resource(data_root,
                            columns=["rating", "review"], target="rating")
        elif self.kind == "subjectivity":
            return Resource(data_root,
                            columns=["subjectivity", "review"],
                            target="subjectivity")
        else:
            return Resource(data_root)

    def _prepare_polarity(self, dataset_root, extracted_path):
        polarity_file_path = os.path.join(dataset_root, "review_polarity.txt")
        negative_path = os.path.join(extracted_path, "txt_sentoken/neg")
        positive_path = os.path.join(extracted_path, "txt_sentoken/pos")

        with open(polarity_file_path, mode="w", encoding="utf-8") as f:
            for i, p in enumerate([negative_path, positive_path]):
                label = i  # negative = 0, positive = 1
                label_name = "negative" if label == 0 else "positive"
                self.logger.info("Extracting {} data.".format(label_name))
                for txt in xtqdm(os.listdir(p)):
                    with open(os.path.join(p, txt), encoding="utf-8") as tf:
                        lines = [ln.strip().replace("\t", " ") for ln in tf.readlines()]
                        review = " ".join(lines)
                        f.write("\t".join([str(label), review]) + "\n")

        return polarity_file_path

    def _prepare_polarity_v1(self, dataset_root, extracted_path):
        polarity_file = os.path.join(dataset_root, "review_polarity_v1.txt")
        with open(polarity_file, mode="w", encoding="utf-8") as f:
            for e in self.extract_targets:
                p = os.path.join(extracted_path, os.path.basename(e))
                label = 0 if e.endswith(".neg") else 1
                label_name = "negative" if label == 0 else "positive"
                self.logger.info("Extracting {} data.".format(label_name))
                total = self.get_line_count(p)
                with open(p, mode="r", errors="replace", encoding="utf-8") as p:
                    for ln in xtqdm(p, total=total):
                        review = ln.strip().replace("\t", " ")
                        f.write("\t".join([str(label), review]) + "\n")

        return polarity_file

    def _prepare_rating(self, dataset_root, extracted_path):
        rating_file_path = os.path.join(dataset_root, "review_rating.txt")
        rating_dir = os.path.join(extracted_path, "scaledata")

        rating_file = open(rating_file_path, "w", encoding="utf-8")
        for user in os.listdir(rating_dir):
            user_dir = os.path.join(rating_dir, user)
            if not os.path.isdir(user_dir):
                continue

            sub_in_review_file = os.path.join(user_dir, "subj." + user)
            user_rating_file = os.path.join(user_dir, "rating." + user)
            total = self.get_line_count(sub_in_review_file)
            self.logger.info("Extracting user {}'s rating data.".format(user))
            with open(sub_in_review_file, "r", encoding="utf-8") as sr:
                with open(user_rating_file, "r", encoding="utf-8") as ur:
                    for review, rating in xtqdm(zip(sr, ur), total=total):
                        _rv = review.strip().replace("\t", " ")
                        _r = rating.strip()
                        rating_file.write("\t".join([_r, _rv]) + "\n")

        rating_file.close()

        return rating_file_path

    def _prepare_subjectivity(self, dataset_root, extracted_path):
        subjectivity_file = os.path.join(dataset_root, "subjectivity.txt")
        with open(subjectivity_file, mode="w", encoding="utf-8") as f:
            for e in self.extract_targets:
                # subjective(plot) = 1
                label = 1 if e.startswith("plot.") else 0
                label_name = "subjective" if label == 1 else "objective"
                self.logger.info("Extracting {} data.".format(label_name))
                p = os.path.join(extracted_path, os.path.basename(e))
                total = self.get_line_count(p)
                with open(p, mode="r", errors="replace", encoding="utf-8") as sb:
                    for ln in xtqdm(sb, total=total):
                        review = ln.strip().replace("\t", " ")
                        f.write("\t".join([str(label), review]) + "\n")

        return subjectivity_file



import os
from chazutsu.datasets.framework.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class DUC2004(Dataset):

    def __init__(self):

        super().__init__(
            name="DUC2004",
            site_url="http://duc.nist.gov/duc2004/tasks.html",
            download_url="https://s3-ap-northeast-1.amazonaws.com/dev.tech-sketch.jp/chakki/chazutsu/DUC2004.zip",  # noqa
            description="DUC2004 Summarization dataset. " \
                        "This dataset has 500 news, and its summary " \
                        "generated by 4 different humans."
            )

    @property
    def extract_targets(self):
        refs = ["DUC2004/task1_ref{}.txt".format(n) for n in range(4)]
        return ["DUC2004/input.txt"] + refs

    def prepare(self, dataset_root, extracted_path):
        paths = self.__get_extracted_pathes(extracted_path)
        write_path = os.path.join(dataset_root, "duc2004.txt")

        data_files = [open(p, encoding="utf-8") for p in paths]
        files_iter = zip(*data_files)
        with open(write_path, "w", encoding="utf-8") as f:
            for lines in files_iter:
                lines = [ln.strip() for ln in lines]
                line = "\t".join(lines) + "\n"
                f.write(line)
            for _f in data_files:
                _f.close()
        return write_path

    def __get_extracted_pathes(self, extracted_path):
        paths = []
        for e in self.extract_targets:
            p = os.path.join(extracted_path, os.path.basename(e))
            paths.append(p)
        return paths

    def make_resource(self, data_root):
        return Resource(
                data_root,
                columns=["news", "summary1", "summary2",
                                 "summary3", "summary4"],
                target="summary1")



import os
import sys
import mmap
import numpy as np
from collections import Counter
from chazutsu.datasets.framework.xtqdm import xtqdm
from chazutsu.datasets.framework.tokenizer import Tokenizer


class Vocabulary():

    def __init__(self, root, name, tokenizer=None,
                 unknown="<unk>", padding="<pad>",
                 end_of_sentence=""):
        self.root = root
        self.name = name
        self._vocab_file_path = os.path.join(root, name + ".vocab")
        self.tokenizer = tokenizer
        if self.tokenizer is None:
            self.tokenizer = Tokenizer()
        self.unknown = unknown
        self.padding = padding
        self.end_of_sentence = end_of_sentence
        self._vocab = {}
        self.__rev_vocab = {}
        self.max_len = 0

        # create logger
        from logging import getLogger, StreamHandler, DEBUG
        self.logger = getLogger(self.name.lower() + "_vocabulary")
        if not self.logger.hasHandlers():
            # logger is global object!
            _level = DEBUG
            handler = StreamHandler(sys.stdout)
            handler.setLevel(_level)
            self.logger.setLevel(_level)
            self.logger.addHandler(handler)

        if self.has_vocab():
            self.load()

    def has_vocab(self):
        return os.path.exists(self._vocab_file_path)

    def __len__(self):
        return len(self._vocab)

    def make(self,
             path_or_paths, vocab_size=-1, min_word_freq=0,
             separator="\t", reserved_words=(), target_column_indexes=()):
        vocab = Counter()
        paths = path_or_paths
        if isinstance(paths, str):
            paths = [paths]

        self.max_len = 0
        for p in paths:
            self.logger.info("Read {} to make vocabulary.".format(p))
            count = self.get_line_count(p)
            for words in xtqdm(self.fetch_line(p, target_column_indexes,
                               separator), total=count):
                for w in words:
                    vocab[w] += 1
                if len(words) > self.max_len:
                    self.max_len = len(words)

        _vocab = [k_v[0] for k_v in vocab.most_common()
                  if not k_v[1] < min_word_freq]
        _rv = reserved_words
        if len(_rv) == 0:
            _rv = [w for w in
                   [self.padding, self.unknown, self.end_of_sentence] if w]
        _vocab = list(_rv) + _vocab

        if vocab_size > 0:
            _vocab = _vocab[:vocab_size]

        self.logger.info(
            "The vocabulary count is {}. You can see it in {}.".format(
                len(_vocab), self._vocab_file_path))
        with open(self._vocab_file_path, "w", encoding="utf-8") as f:
            f.write("\n".join(_vocab))
        self._vocab = dict(zip(_vocab, range(len(_vocab))))
        self.__rev_vocab = {}

    def get_line_count(self, file_path):
        count = 0
        with open(file_path, "r+") as fp:
            buf = mmap.mmap(fp.fileno(), 0)
            while buf.readline():
                count += 1
        return count

    def fetch_line(self, path, target_column_indexes=(), separator="\t"):
        with open(path, encoding="utf-8") as f:
            for line in f:
                _line = line
                if len(target_column_indexes) > 0:
                    els = _line.split(separator)
                    els = [els[i] for i in target_column_indexes]
                    _line = " ".join(els)

                words = self.tokenizer.tokenize(_line)
                if self.end_of_sentence:
                    words.append(self.end_of_sentence)

                yield words

    def load(self):
        if self.has_vocab():
            self.__rev_vocab = {}
            with open(self._vocab_file_path, encoding="utf-8") as f:
                lines = f.readlines()
                for i, ln in enumerate(lines):
                    w = ln.strip()
                    self._vocab[w] = i

    def str_to_ids(self, sentence):
        if len(self._vocab) == 0:
            self.load()
        _s = sentence
        if self.end_of_sentence:
            _s = _s.replace("\r\n", self.end_of_sentence)
            _s = _s.replace("\n", self.end_of_sentence)
        words = self.tokenizer.tokenize(sentence)
        unk_id = self._vocab[self.unknown]
        ids = [unk_id if w not in self._vocab else self._vocab[w]
               for w in words]
        return ids

    def ids_to_words(self, ids, ignore_padding=False):
        if len(self._vocab) == 0:
            self.load()

        if len(self.__rev_vocab) == 0:
            self.__rev_vocab = {v: k for k, v in self._vocab.items()}

        words = [self.__rev_vocab[i] for i in ids]
        if ignore_padding:
            words = [w for w in words if w != self.padding]
        return words

    def str_to_matrix(self, sentence, fixed_len=-1):
        ids = self.str_to_ids(sentence)
        pad_id = self._vocab[self.padding]
        if fixed_len > 0:
            if len(ids) > fixed_len:
                ids = ids[:fixed_len]
            elif len(ids) < fixed_len:
                ids = ids + ([pad_id] * (fixed_len - len(ids)))

        seqlen_x_vocab = np.zeros((len(ids), len(self._vocab)))
        for i, _id in enumerate(ids):
            seqlen_x_vocab[i][_id] = 1
        return seqlen_x_vocab

    def matrix_to_words(self, matrix, ignore_padding=False):
        ids = np.argmax(matrix, axis=1)
        return self.ids_to_words(ids, ignore_padding)



from setuptools import setup


setup(
    name="chazutsu",
    version="0.8.2",
    description="The tool to make NLP datasets ready to use",
    keywords=["machine learning", "nlp", "natural language processing"],
    author="icoxfog417",
    author_email="icoxfog417@yahoo.co.jp",
    license="Apache License 2.0",
    packages=[
        "chazutsu",
        "chazutsu.datasets",
        "chazutsu.datasets.framework"
        ],
    url="https://github.com/chakki-works/chazutsu",
    install_requires=[
        "requests>=2.14.2",
        "tqdm>=4.11.2",
        "pandas>=0.20.1",
        "joblib>=0.11",
    ],
)



import os
import mmap
import numpy as np
import pandas as pd
from chazutsu.datasets.framework.vocabulary import Vocabulary
from chazutsu.datasets.framework import converter as cv


class Resource():

    def __init__(self,
                 root,
                 columns=None, target="",
                 separator="\t", pattern=()):

        self.root = root
        self.columns = columns
        self.target = target
        self.separator = separator
        self._pattern = pattern
        self._vocab = None
        if len(self._pattern) == 0:
            self._pattern = {
                "train": "_train",
                "test": "_test",
                "valid": "_valid",
                "sample": "_samples"
            }
        self._resource_name = ""
        self._resource = {}
        self._batch_def = {}

        if target and target not in columns:
            raise Exception("Target have to be selected from columns.")
        self.find_resource()

    def find_resource(self):
        for f in os.listdir(self.root):
            p = os.path.join(self.root, f)
            if not os.path.isfile(p) or f.startswith("."):
                continue

            n, e = os.path.splitext(f)
            if e == ".vocab":
                continue  # skip vocab file

            match = False
            for kind in self._pattern:
                if n.endswith(self._pattern[kind]):
                    self._resource[kind] = p
                    self._resource_name = n.replace(self._pattern[kind], "")
                    match = True
                    break

            if not match:
                self._resource["data"] = p
                self._resource_name = n

        if "data" not in self._resource and "train" in self._resource:
            self._resource["data"] = self._resource["train"]

    @property
    def data_file_path(self):
        return self._get_prop("data")

    @property
    def train_file_path(self):
        return self._get_prop("train")

    @property
    def test_file_path(self):
        return self._get_prop("test")

    @property
    def valid_file_path(self):
        return self._get_prop("valid")

    @property
    def sample_file_path(self):
        return self._get_prop("sample")

    def _get_prop(self, name):
        return "" if name not in self._resource else self._resource[name]

    def data(self, split_target=False):
        return self._get_data("data", split_target)

    def train_data(self, split_target=False):
        return self._get_data("train", split_target)

    def test_data(self, split_target=False):
        return self._get_data("test", split_target)

    def valid_data(self, split_target=False):
        return self._get_data("valid", split_target)

    def sample_data(self, split_target=False):
        return self._get_data("sample", split_target)

    def _get_data(self, kind, split_target):
        if kind not in self._resource:
            raise Exception("Can not find {} data self.".format(kind))
        return self._to_pandas(self._resource[kind], split_target)

    def _to_pandas(self, path, split_target):
        df = pd.read_csv(path, header=None, names=self.columns, sep="\t")

        if not split_target:
            return df
        elif self.target and self.target in df.columns:
            target = df[self.target]
            df.drop(self.target, axis=1, inplace=True)
            return target, df
        else:
            return df

    def make_vocab(self,
                   vocab_resources=("train", "valid", "test"),
                   columns_for_vocab=(),
                   tokenizer=None,
                   vocab_size=-1,
                   min_word_freq=0,
                   unknown="<unk>",
                   padding="<pad>",
                   end_of_sentence="",
                   reserved_words=("<pad>", "<unk>"),
                   force=False):

        if len(self._resource) == 0:
            raise Exception("Supplied resource does not have any resource.")

        # Prepare the file paths to make vocab
        resource_for_vocab = []
        for kind in vocab_resources:
            if kind in self._resource and self._resource[kind]:
                resource_for_vocab.append(self._resource[kind])
        if len(resource_for_vocab) == 0:
            raise Exception(
                "There are no resource in {} to make vocaburaly.".format(
                    self._resource_name))

        # Prepare the column index to make vocab
        _c_for_vocab = columns_for_vocab
        if len(_c_for_vocab) == 0:
            _c_for_vocab = self.columns  # all columns is used

        _c_for_vocab_idx = [i for i, c in enumerate(self.columns)
                            if c in _c_for_vocab]

        self._vocab = Vocabulary(
            self.root, self._resource_name,
            tokenizer=tokenizer, unknown=unknown, padding=padding,
            end_of_sentence=end_of_sentence)

        if self._vocab.has_vocab() and not force:
            self._vocab.load()
        else:
            self._vocab.make(
                path_or_paths=resource_for_vocab,
                vocab_size=vocab_size, min_word_freq=min_word_freq,
                separator=self.separator, reserved_words=reserved_words,
                target_column_indexes=_c_for_vocab_idx)
        return self._vocab

    @property
    def vocab_file_path(self):
        if self._vocab.has_vocab():
            return self._vocab._vocab_file_path
        else:
            return ""

    @property
    def vocab(self):
        v = self._vocab._vocab
        return v

    def column(self, column):
        return Route(self, column)

    def to_batch(self, kind, columns=()):
        df = self._get_data(kind, split_target=False)
        y = None
        Xs = []
        _columns = columns if len(columns) > 0 else self.columns
        for c in _columns:
            if c not in df.columns:
                raise Exception("{} is not exist.".format(c))
            s = self._to_array(c, df[c])
            if c == self.target:
                y = s
            else:
                Xs.append(s)
        if len(Xs) == 1:
            X = Xs[0]
        else:
            X = np.hstack(Xs)

        return X, y

    def _to_array(self, column, data):
        _cv = cv.Converter()
        if column in self._batch_def:
            _cv = self._batch_def[column]

        s = _cv.flow(data)
        if len(s.shape) == 1:
            s = np.reshape(s, (-1, 1))  # vector to vertical mx
        return s

    def get_line_count(self, kind):
        file_path = self._get_prop(kind)
        count = 0
        with open(file_path, "r+") as fp:
            buf = mmap.mmap(fp.fileno(), 0)
            while buf.readline():
                count += 1
        return count

    def to_batch_iter(self, kind, batch_size, columns=()):
        iterator = self._to_batch_iter(kind, batch_size, columns)
        batch_count = self.get_line_count(kind) // batch_size
        return iterator, batch_count

    def _to_batch_iter(self, kind, batch_size, columns=()):
        file_path = self._get_prop(kind)
        targets = columns if len(columns) > 0 else self.columns
        while True:
            Xs = {}
            y = []
            with open(file_path, encoding="utf-8") as f:
                for line in f:
                    tokens = line.strip().split(self.separator)
                    for i, c in enumerate(self.columns):
                        if c not in targets:
                            continue
                        if c == self.target:
                            y.append(tokens[i])
                        else:
                            if c not in Xs:
                                Xs[c] = []
                            Xs[c].append(tokens[i])
                    if len(y) == batch_size:
                        yb = self._to_array(self.target, y)
                        Xb = []
                        for c in Xs:
                            Xb.append(self._to_array(c, Xs[c]))

                        if len(Xb) == 1:
                            Xb = Xb[0]
                        else:
                            Xb = np.hstack(Xb)
                        yield Xb, yb

                        y.clear()
                        for x in Xs:
                            Xs[x].clear()


class Route():

    def __init__(self, resource, column):
        self.r = resource
        self.column = column

    def as_word_seq(self, fixed_len=-1):
        c = cv.VocabularyConverter(self.r._vocab, fixed_len)
        self.r._batch_def[self.column] = c
        return self

    def as_category(self, labels=(), distincts=False, num_class=-1):
        c = cv.OneHotConverter(labels, distincts, num_class)
        self.r._batch_def[self.column] = c
        return self

    def as_dictid(self, labels=(), distincts=False, num_class=-1):
        c = cv.DictionalyConverter(labels, distincts, num_class)
        self.r._batch_def[self.column] = c
        return self

    def flow(self, series):
        if self.column in self.r._batch_def:
            return self.r._batch_def[self.column].flow(series)
        else:
            return None

    def back(self, converted):
        if self.column in self.r._batch_def:
            return self.r._batch_def[self.column].back(converted)
        else:
            return None

    def to_batch(self, kind, with_target=False):
        if with_target:
            return self.r.to_batch(kind, columns=[self.r.target, self.column])
        else:
            X, _ = self.r.to_batch(kind, columns=[self.column])
            return X
